@@ -24,8 +24,11 @@
  *      Jacob Berkman <jacob@helixcode.com>
  */
 
+#include <config.h>
+
 #include <gtk/gtk.h>
 
+#include <glib/gi18n-lib.h>
 #include "selector.h"
 #include "inlinepixbufs.h"
 #include "libwnck.h"
