@@ -21,9 +21,13 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
+
+#include <glib/gi18n-lib.h>
 #include <string.h>
 #include "window.h"
 #include "class-group.h"
+#include "util.h"
 #include "xutils.h"
 #include "private.h"
 #include "wnck-enum-types.h"
@@ -48,7 +52,8 @@ static GHashTable *window_hash = NULL;
     ((window)->priv->is_hidden           << 7) |        \
     ((window)->priv->is_fullscreen       << 8) |        \
     ((window)->priv->demands_attention   << 9) |        \
-    ((window)->priv->is_urgent           << 10) )
+    ((window)->priv->is_urgent           << 10)|        \
+    ((window)->priv->is_above            << 11))
 
 struct _WnckWindowPrivate
 {
@@ -58,6 +63,7 @@ struct _WnckWindowPrivate
   WnckClassGroup *class_group;
   Window group_leader;
   Window transient_for;
+  GdkRectangle icon_geometry;
   char *name;
   char *icon_name;
   char *session_id;
@@ -95,6 +101,7 @@ struct _WnckWindowPrivate
   guint is_maximized_horz : 1;
   guint is_maximized_vert : 1;
   guint is_shaded : 1;
+  guint is_above : 1;
   guint skip_pager : 1;
   guint skip_taskbar : 1;
   guint is_sticky : 1;
@@ -121,13 +128,15 @@ struct _WnckWindowPrivate
   guint need_update_wm_state : 1;
   guint need_update_icon_name : 1;
   guint need_update_workspace : 1;
-  guint need_emit_icon_changed : 1;
   guint need_update_actions : 1;
   guint need_update_wintype : 1;
   guint need_update_transient_for : 1;
   guint need_update_startup_id : 1;
   guint need_update_wmclass : 1;
   guint need_update_wmhints : 1;
+
+  guint need_emit_name_changed : 1;
+  guint need_emit_icon_changed : 1;
 };
 
 enum {
@@ -211,9 +220,10 @@ wnck_window_init (WnckWindow *window)
 {
   window->priv = g_new0 (WnckWindowPrivate, 1);
 
-  window->priv->name = g_strdup (FALLBACK_NAME);
+  window->priv->name = NULL;
   window->priv->icon_name = NULL;
   window->priv->workspace = ALL_WORKSPACES;
+  window->priv->icon_geometry.width = -1; /* invalid cached value */
 
   window->priv->icon_cache = _wnck_icon_cache_new ();
 }
@@ -411,6 +421,8 @@ _wnck_window_create (Window      xwindow,
   window->priv->need_update_startup_id = TRUE;
   window->priv->need_update_wmclass = TRUE;
   window->priv->need_update_wmhints = TRUE;
+  window->priv->need_emit_name_changed = FALSE;
+  window->priv->need_emit_icon_changed = FALSE;
   force_update_now (window);
 
   return window;
@@ -431,13 +443,36 @@ _wnck_window_destroy (WnckWindow *window)
   g_object_unref (G_OBJECT (window));
 }
 
+/**
+ * wnck_window_has_name:
+ * @window: a #WnckWindow
+ *
+ * Checks whether or not @window has a name. wnck_window_get_name()
+ * will always return some value, even if the window hasn't set a
+ * name; wnck_window_has_name() can be used to tell if that name is
+ * real or not.
+ *
+ * For icons titles, use wnck_window_has_icon_name() instead.
+ *
+ * Return value: %TRUE if wnck_window_get_name() returns the window's
+ * real name, %FALSE if it returns a fallback name.
+ **/
+gboolean
+wnck_window_has_name (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
+
+  return window->priv->name != NULL;
+}
+
 /**
  * wnck_window_get_name:
  * @window: a #WnckWindow
  *
  * Gets the name of the window, as it should be displayed in a pager
  * or tasklist. Always returns some value, even if the window
- * hasn't set a name.
+ * hasn't set a name; use wnck_window_has_name() if you need to know
+ * whether the returned name is "real" or not.
  *
  * For icons titles, use wnck_window_get_icon_name() instead.
  *
@@ -448,17 +483,47 @@ wnck_window_get_name (WnckWindow *window)
 {
   g_return_val_if_fail (WNCK_IS_WINDOW (window), NULL);
 
-  return window->priv->name;
+  if (window->priv->name)
+    return window->priv->name;
+  else
+    return FALLBACK_NAME;
+}
+
+/**
+ * wnck_window_has_icon_name:
+ * @window: a #WnckWindow
+ *
+ * Checks whether or not @window has an icon name.
+ * wnck_window_get_icon_name() will always return some value, even if
+ * the window hasn't set an icon name; wnck_window_has_icon_name() can
+ * be used to tell if that name is real or not.
+ *
+ * (Note that if wnck_window_has_icon_name() returns %FALSE, but
+ * wnck_window_has_name() returns %TRUE, then the name returned by
+ * wnck_window_get_icon_name() is the window name. Only when both
+ * methods return %FALSE does wnck_window_get_icon_name() return a
+ * generic fallback name.)
+ *
+ * Return value: %TRUE if wnck_window_get_icon_name() returns the
+ * window's real icon name, %FALSE if it returns a fallback name.
+ **/
+gboolean
+wnck_window_has_icon_name (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
+
+  return window->priv->icon_name != NULL;
 }
 
 /**
  * wnck_window_get_icon_name:
  * @window: a #WnckWindow
  *
- * Gets the name of the window, as it should be displayed for an
- * icon. Always returns some value, even if the window hasn't set a
- * name. Contrast with wnck_window_get_name(), which returns the
- * window title, not the icon title.
+ * Gets the name of the window, as it should be displayed for an icon.
+ * Always returns some value, even if the window hasn't set a name;
+ * use wnck_window_has_icon_name() if you need to know if the name is
+ * "real" or not. Contrast with wnck_window_get_name(), which returns
+ * the window title, not the icon title.
  *
  * Return value: name of the window
  **/
@@ -469,8 +534,10 @@ wnck_window_get_icon_name (WnckWindow *window)
 
   if (window->priv->icon_name)
     return window->priv->icon_name;
-  else
+  else if (window->priv->name)
     return window->priv->name;
+  else
+    return FALLBACK_NAME;
 }
 
 WnckApplication*
@@ -823,6 +890,14 @@ wnck_window_is_shaded                 (WnckWindow *window)
   return window->priv->is_shaded;
 }
 
+gboolean
+wnck_window_is_above                  (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
+
+  return window->priv->is_above;
+}
+
 gboolean
 wnck_window_is_skip_pager             (WnckWindow *window)
 {
@@ -1026,6 +1101,30 @@ wnck_window_unshade                 (WnckWindow *window)
                       0);
 }
 
+void
+wnck_window_make_above (WnckWindow *window)
+{
+  g_return_if_fail (WNCK_IS_WINDOW (window));
+
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+                      window->priv->xwindow,
+                      TRUE,
+                      _wnck_atom_get ("_NET_WM_STATE_ABOVE"),
+                      0);
+}
+
+void
+wnck_window_unmake_above (WnckWindow *window)
+{
+  g_return_if_fail (WNCK_IS_WINDOW (window));
+
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+                      window->priv->xwindow,
+                      FALSE,
+                      _wnck_atom_get ("_NET_WM_STATE_ABOVE"),
+                      0);
+}
+
 void
 wnck_window_stick                   (WnckWindow *window)
 {
@@ -1483,6 +1582,43 @@ wnck_window_get_geometry (WnckWindow *window,
     *heightp = window->priv->height;
 }
 
+/**
+ * wnck_window_set_geometry:
+ * @window: a #WnckWindow
+ * @gravity: one of #WnckWindowGravity
+ * @geometry_mask: a mask of #WnckWindowMoveResizeMask
+ * @x: new X coordinate of window
+ * @y: new Y coordinate of window
+ * @width: new width of window
+ * @height: new height of window
+ *
+ * Sets the size and position of the window.
+ *
+ **/
+void
+wnck_window_set_geometry (WnckWindow               *window,
+                          WnckWindowGravity         gravity,
+                          WnckWindowMoveResizeMask  geometry_mask,
+                          int                       x,
+                          int                       y,
+                          int                       width,
+                          int                       height)
+{
+  int gravity_and_flags;
+  int source;
+
+  g_return_if_fail (WNCK_IS_WINDOW (window));
+
+  source = _wnck_get_client_type();
+  gravity_and_flags = gravity;
+  gravity_and_flags |= geometry_mask << 8;
+  gravity_and_flags |= source << 12;
+
+  _wnck_set_window_geometry (WNCK_SCREEN_XSCREEN (window->priv->screen),
+                             window->priv->xwindow,
+                             gravity_and_flags, x, y, width, height);
+}
+
 /**
  * wnck_window_is_visible_on_workspace:
  * @window: a #WnckWindow
@@ -1527,6 +1663,17 @@ wnck_window_set_icon_geometry (WnckWindow *window,
 			       int         width,
 			       int         height)
 {
+  if (window->priv->icon_geometry.x == x &&
+      window->priv->icon_geometry.y == y &&
+      window->priv->icon_geometry.width == width &&
+      window->priv->icon_geometry.height == height)
+    return;
+
+  window->priv->icon_geometry.x = x;
+  window->priv->icon_geometry.y = y;
+  window->priv->icon_geometry.width = width;
+  window->priv->icon_geometry.height = height;
+
   _wnck_set_icon_geometry (window->priv->xwindow,
                            x, y, width, height);
 }
@@ -1773,9 +1920,11 @@ update_state (WnckWindow *window)
       window->priv->is_maximized_vert = FALSE;
       window->priv->is_sticky = FALSE;
       window->priv->is_shaded = FALSE;
+      window->priv->is_above = FALSE;
       window->priv->skip_taskbar = FALSE;
       window->priv->skip_pager = FALSE;
       window->priv->net_wm_state_hidden = FALSE;
+      window->priv->is_fullscreen = FALSE;
       window->priv->demands_attention = FALSE;
       
       atoms = NULL;
@@ -1797,6 +1946,8 @@ update_state (WnckWindow *window)
             window->priv->is_sticky = TRUE;
           else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_SHADED"))
             window->priv->is_shaded = TRUE;
+          else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_ABOVE"))
+            window->priv->is_above = TRUE;
           else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_FULLSCREEN"))
             window->priv->is_fullscreen = TRUE;
           else if (atoms[i] == _wnck_atom_get ("_NET_WM_STATE_SKIP_TASKBAR"))
@@ -1877,48 +2028,61 @@ update_state (WnckWindow *window)
     }
 }
 
+static gboolean
+nullstr_equal (const char *str1, const char *str2)
+{
+  if (str1 == NULL || str2 == NULL)
+    return str1 == str2;
+  else
+    return !strcmp (str1, str2);
+}
+
 static void
 update_name (WnckWindow *window)
 {
-  g_return_if_fail (window->priv->name == NULL);
+  char *new_name;
 
   if (!window->priv->need_update_name)
     return;
 
   window->priv->need_update_name = FALSE;
 
-  window->priv->name = _wnck_get_name (window->priv->xwindow);
+  new_name = _wnck_get_name (window->priv->xwindow);
 
-  if (window->priv->name == NULL)
-    window->priv->name = g_strdup (FALLBACK_NAME);
+  if (!nullstr_equal (window->priv->name, new_name))
+    window->priv->need_emit_name_changed = TRUE;
+
+  g_free (window->priv->name);
+  window->priv->name = new_name;
 }
 
 static void
 update_icon_name (WnckWindow *window)
 {
-  g_return_if_fail (window->priv->icon_name == NULL);
+  char *new_name = NULL;
 
   if (!window->priv->need_update_icon_name)
     return;
 
   window->priv->need_update_icon_name = FALSE;
 
-  if (window->priv->icon_name == NULL)
-    window->priv->icon_name =
-      _wnck_get_utf8_property (window->priv->xwindow,
-                               _wnck_atom_get ("_NET_WM_VISIBLE_ICON_NAME"));
+  if (new_name == NULL)
+    new_name = _wnck_get_utf8_property (window->priv->xwindow,
+                                        _wnck_atom_get ("_NET_WM_VISIBLE_ICON_NAME"));
 
-  if (window->priv->icon_name == NULL)
-    window->priv->icon_name =
-      _wnck_get_utf8_property (window->priv->xwindow,
-                               _wnck_atom_get ("_NET_WM_ICON_NAME"));
+  if (new_name == NULL)
+    new_name = _wnck_get_utf8_property (window->priv->xwindow,
+                                        _wnck_atom_get ("_NET_WM_ICON_NAME"));
 
-  if (window->priv->icon_name == NULL)
-    window->priv->icon_name =
-      _wnck_get_text_property (window->priv->xwindow,
-                               XA_WM_ICON_NAME);
+  if (new_name == NULL)
+    new_name = _wnck_get_text_property (window->priv->xwindow,
+                                        XA_WM_ICON_NAME);
 
-  /* no fallback, get_icon_name falls back to regular window title */
+  if (!nullstr_equal (window->priv->icon_name, new_name))
+    window->priv->need_emit_name_changed = TRUE;
+
+  g_free (window->priv->icon_name);
+  window->priv->icon_name = new_name;
 }
 
 static void
@@ -1968,6 +2132,7 @@ update_actions (WnckWindow *window)
                 WNCK_WINDOW_ACTION_MOVE                    |
                 WNCK_WINDOW_ACTION_RESIZE                  |
                 WNCK_WINDOW_ACTION_SHADE                   |
+                WNCK_WINDOW_ACTION_ABOVE                   |
                 WNCK_WINDOW_ACTION_STICK                   |
                 WNCK_WINDOW_ACTION_MAXIMIZE_HORIZONTALLY   |
                 WNCK_WINDOW_ACTION_MAXIMIZE_VERTICALLY     |
@@ -2126,7 +2291,8 @@ update_transient_for (WnckWindow *window)
   
   if (_wnck_get_window (window->priv->xwindow,
                         _wnck_atom_get ("WM_TRANSIENT_FOR"),
-                        &parent))
+                        &parent) &&
+      parent != window->priv->xwindow)
     {
       window->priv->transient_for = parent;
 
@@ -2211,9 +2377,6 @@ force_update_now (WnckWindow *window)
   WnckWindowState old_state;
   WnckWindowState new_state;
   WnckWindowActions old_actions;
-  char *old_name;
-  char *old_icon_name;
-  gboolean do_emit_name_changed = FALSE;
   
   unqueue_update (window);
 
@@ -2223,36 +2386,10 @@ force_update_now (WnckWindow *window)
    * and we have to fix that before we emit any other signals
    */
 
-  old_name = window->priv->name;
-  window->priv->name = NULL;
-
   update_name (window);
-
-  if (window->priv->name == NULL)
-    window->priv->name = old_name;
-  else
-    {
-      if (strcmp (window->priv->name, old_name) != 0)
-        do_emit_name_changed = TRUE;
-      g_free (old_name);
-    }
-
-  old_icon_name = window->priv->icon_name;
-  window->priv->icon_name = NULL;
-
   update_icon_name (window);
 
-  if (window->priv->icon_name == NULL)
-    window->priv->icon_name = old_icon_name;
-  else
-    {
-      if (old_icon_name == NULL ||
-          strcmp (window->priv->icon_name, old_icon_name) != 0)
-        do_emit_name_changed = TRUE;
-      g_free (old_icon_name);
-    }
-
-  if (do_emit_name_changed)
+  if (window->priv->need_emit_name_changed)
     emit_name_changed (window);
 
   old_state = COMPRESS_STATE (window);
@@ -2318,6 +2455,7 @@ unqueue_update (WnckWindow *window)
 static void
 emit_name_changed (WnckWindow *window)
 {
+  window->priv->need_emit_name_changed = FALSE;
   g_signal_emit (G_OBJECT (window),
                  signals[NAME_CHANGED],
                  0);
